from __future__ import print_function

from credo.errors import BadCredentialFile, NoCredentialsFound, BadCredential, UserQuit
from credo.asker import ask_user_for_half_life, ask_for_choice_or_new
from credo.amazon import IamPair

import logging
import sys

log = logging.getLogger("credo.cred_types.amazon")

class AmazonKey(object):
    """Represents the information and meta information required for amazon credentials"""
    def __init__(self, key_info, credential_path, iam_pair=None, iam_pairs=None):
        self.key_info = key_info
        self.credential_path = credential_path

        self.iam_pairs = iam_pairs
        if self.iam_pairs is None:
            self.iam_pairs = {}

        if iam_pair is not None:
            pair = (iam_pair.aws_access_key_id, iam_pair.aws_secret_access_key)
            if pair not in self.iam_pairs:
                self.iam_pairs[pair] = iam_pair
            self._decrypted = [pair]

    @property
    def crypto(self):
        """Proxy credential_path"""
        return self.credential_path.crypto

    @property
    def changed(self):
        """Say whether this key has changed"""
        if self.iam_pair:
            return self.iam_pair.changed
        else:
            return False

    def unchanged(self):
        """Set the key as unchanged"""
        self._changed = False
        if self.iam_pair:
            self.iam_pair.unchanged()

    def mark_as_invalid(self):
        """Mark our key as invalid"""
        if self.iam_pair:
            self.iam_pair.mark_as_invalid()

    @classmethod
    def using(kls, iam_pair, credential_path, iam_pairs):
        """Create an AmazonKey from the provided details"""
        if not iam_pair.works:
            raise BadCredential()

        key_info = {"fingerprints": None, "create_epoch": iam_pair.create_epoch, "half_life": iam_pair.half_life}
        key = AmazonKey(key_info, credential_path, iam_pair, iam_pairs=iam_pairs)
        if not key.verifier(data=None, iam_pair=iam_pair):
            raise BadCredential()
        return key

    def basic_validation(self):
        """Make sure the keys have basic requirements"""
        if "fingerprints" not in self.key_info:
            return "No fingerprints for this key"
        if not isinstance(self.fingerprints, dict):
            return "Fingerprints for this key are not a dictionary"
        if not self.crypto.decryptable(self.fingerprints):
            return "No private key can decrypt secrets"
        if any(not all(attr in value for attr in ("secret", "data", "verifier")) for value in self.fingerprints.values()):
            return "One or more of the fingeprints doesn't contain secret, data and verifier"

    @property
    def fingerprints(self):
        """Get our fingerprints from key_info"""
        if "fingerprints" not in self.key_info:
            self.key_info["fingerprints"] = {}
        return self.key_info["fingerprints"]

    def credentials(self):
        """Goes through our fingerprints and yields all our decryptable credentials as [aws_access_key_id, aws_secret_access_key]"""
        if getattr(self, "_decrypted", None):
            for key in self._decrypted:
                yield key
            return

        decrypted = self.crypto.decrypt_by_fingerprint(self.fingerprints, self.verifier)
        if decrypted:
            yield decrypted["aws_access_key_id"], decrypted["aws_secret_access_key"]

    @property
    def iam_pair(self):
        """Find the first verified iam pair"""
        if not getattr(self, "_iam_pair", None):
            for aws_access_key_id, aws_secret_access_key in self.credentials():
                self._iam_pair = self.make_iam_pair(aws_access_key_id, aws_secret_access_key)
                break
        return getattr(self, "_iam_pair", None)

    def make_iam_pair(self, access_key, secret_key, half_life=None):
        """Make an iam pair, or get cached pair"""
        if (access_key, secret_key) not in self.iam_pairs:
            if half_life is None:
                half_life = self.key_info.get("half_life")
            iam_pair = IamPair(access_key, secret_key, create_epoch=self.key_info.get("create_epoch"), half_life=half_life)
            self.iam_pairs[(access_key, secret_key)] = iam_pair
        return self.iam_pairs[(access_key, secret_key)]

    @property
    def encrypted_values(self):
        """
        Return this key as a dictionary of {"fingerprints": {<fingerprint>: <info>, ...}, <other_options>}

        Where <info> is {"aws_access_key_id", "aws_secret_access_key"}

        and <other_options> includes {"create_epoch"}
        """
        fingerprints = self.crypto.fingerprinted({"aws_access_key_id": self.iam_pair.aws_access_key_id, "aws_secret_access_key": self.iam_pair.aws_secret_access_key})
        return {"fingerprints": fingerprints, "create_epoch": self.iam_pair.create_epoch, "half_life": self.iam_pair.half_life}

    def verifier(self, data=None, iam_pair=None):
        """Say that these values is an amazon key for this account and user"""
        if data is not None:
            iam_pair = self.make_iam_pair(data.get("aws_access_key_id"), data.get("aws_secret_access_key"))

        if not iam_pair.works:
            return False

        amazon_account = iam_pair.ask_amazon_for_account()
        amazon_username = iam_pair.ask_amazon_for_username()

        # Get the account and username they should be
        account = self.credential_path.account.account_id(iam_pair=iam_pair)
        username = self.credential_path.user.username(iam_pair=iam_pair)

        if account != amazon_account or username != amazon_username:
            log.error("Expected key with different account and username\tgot_account=%s\texpected_account=%s\tgot_username=%s\texpected_username=%s"
                , amazon_account, account, amazon_username, username
                )
            return False
        else:
            return True

class AmazonKeys(object):
    """Collection of Amazon keys"""
    type = "amazon"

    def __init__(self, keys, credential_path):
        if not keys:
            keys = []

        self.credential_path = credential_path
        self.iam_pairs = {}
        self.keys = [AmazonKey(key, credential_path, iam_pairs=self.iam_pairs) for key in keys]
        self._changed = False

    @property
    def crypto(self):
        """Proxy credential_path"""
        return self.credential_path.crypto

    def __iter__(self):
        """Iterate through the keys"""
        return iter(self.keys)

    def __len__(self):
        """Return how many keys we have"""
        return len(self.keys)

    def add(self, iam_pair):
        """Add a key"""
        key = AmazonKey.using(iam_pair, self.credential_path, iam_pairs=self.iam_pairs)
        if not key.iam_pair or not key.iam_pair.works:
            raise BadCredential()

        found = False
        for other in self.keys:
            if other.iam_pair and key.iam_pair:
                if other.iam_pair.aws_access_key_id == key.iam_pair.aws_access_key_id:
                    other.iam_pair.synchronize_with(key.iam_pair)
                    found = True

        if not found:
            self.keys.append(key)
            self._changed = True

    @property
    def changed(self):
        """Say whether there has been any changes"""
        return self._changed or any(key.changed for key in self.keys)

    def unchanged(self):
        """Reset changed on everything"""
        self._changed = False
        for key in self.keys:
            key.unchanged()

    def invalidate_all(self):
        """Mark all the keys as invalid"""
        for key in self.keys:
            key.mark_as_invalid()

    @property
    def iam_pair(self):
        """Find the youngest iam pair"""
        keys = list(key for key in self.keys if key.iam_pair and key.iam_pair.works)
        if keys:
            return sorted(keys, key=lambda k: k.iam_pair.create_epoch, reverse=True)[0].iam_pair

    def basic_keys_validation(self):
        """Do some basic validation of our keys"""
        errors = []
        for index, key in enumerate(self.keys):
            nxt = key.basic_validation()
            if nxt:
                errors.append("key {0}: {1}".format(index+1, nxt))

        if errors:
            raise BadCredentialFile("Some of the keys were not valid", errors=errors)

    @property
    def encrypted_values(self):
        """Return our keys as a dictionary with encrypted values"""
        result = []
        access_keys = []
        log.info("Making encrypted values for some keys")
        for key in self.keys:
            if key.iam_pair and key.iam_pair.works and not key.iam_pair.deleted:
                result.append(key.encrypted_values)
                access_keys.append(key.iam_pair.aws_access_key_id)
            else:
                access_key=""
                if key.iam_pair:
                    access_key="\taccess_key={0}".format(list(key.credentials())[0][0])
                log.info("Not saving invalid credentials%s", access_key)

        log.info("Made encrypted values for %s keys using %s public keys", len(result), len(self.crypto.public_key_fingerprints))
        return result, access_keys

    @property
    def access_keys(self):
        """Return all the access keys we know about"""
        for key in self.keys:
            iam_pair = key.iam_pair
            if iam_pair and iam_pair.works:
                yield iam_pair.aws_access_key_id

    def exports(self):
        """Return a list of (key, value) exports that would be necessary for the main iam_pair"""
        iam_pair = self.iam_pair
        if not iam_pair:
            raise NoCredentialsFound()

        return [
              ("AWS_ACCESS_KEY_ID", iam_pair.aws_access_key_id)
            , ("AWS_SECRET_ACCESS_KEY", iam_pair.aws_secret_access_key)
            ]

    def add_key(self, iam_pair):
        """Create a new key and add to our collection"""
        key = AmazonKey.using(iam_pair, self.credential_path, self.iam_pairs)
        self.keys.append(key)
        self._changed = True
        return key

    def deal_with_unknown_key(self, access_key, valid_iam_pair):
        """Work out what to do with this key credo doesn't know about"""
        while True:
            quit_choice = "Quit"
            delete_choice = "Delete key"
            choice = ask_for_choice_or_new(
                  "action for dealing with an access_key credo doesn't know about ({0})".format(access_key)
                , [quit_choice, delete_choice]
                )

            if choice == quit_choice:
                raise UserQuit
            elif choice == delete_choice:
                valid_iam_pair.delete_access_key(access_key)
                return
            else:
                iam_pair = IamPair(access_key, choice)
                if iam_pair.works:
                    half_life = ask_user_for_half_life(iam_pair.aws_access_key_id)
                    iam_pair.set_half_life(half_life)
                    key = self.add_key(iam_pair)
                    if key:
                        return key
                else:
                    print("The secret key you entered was not valid", file=sys.stderr)

    def needs_rotation(self):
        """Say whether the current keys that we know about need any rotation"""
        working = []
        for key in self.keys:
            if key.iam_pair and key.iam_pair.works:
                working.append(key)
                if key.iam_pair.past_half_life() or key.iam_pair.expired():
                    return True

        # Only need rotation if we have no working keys
        return len(working) == 0

    def rotate(self, half_life=None):
        """Rotate the keys and return whether any of them changed"""
        while True:
            # Keep looking at the keys until we have no more surprises
            counts = {"created": 0, "deleted": 0, "removed": 0, "resolved": 0}
            usable = []
            to_remain = []
            for_deletion = []
            for_rotation = []

            known = []

            for key in self.keys:
                if not key.iam_pair or not key.iam_pair.works:
                    counts["removed"] += 1
                else:
                    usable.append(key.iam_pair)
                    known.append(key.iam_pair.aws_access_key_id)
                    if key.iam_pair.expired():
                        counts["deleted"] += 1
                        for_deletion.append(key.iam_pair)
                    elif key.iam_pair.past_half_life():
                        counts["created"] += 1
                        for_rotation.append(key.iam_pair)
                        to_remain.append(key.iam_pair)
                    else:
                        to_remain.append(key.iam_pair)

            if not to_remain:
                counts["created"] += 1

            extras = []
            if any(usable) and not usable[0].is_root_credentials():
                others = usable[0].find_other_access_keys()
                for access_key in others:
                    if access_key not in known:
                        new_key = self.deal_with_unknown_key(access_key, usable[0])
                        counts["resolved"] += 1
                        if new_key:
                            extras.append(new_key)

            if not extras:
                break

        if not any(counts.values()):
            return False
        else:
            self._changed = True

        log.info("Rotation will result in creating %s keys, deleting %s keys, removing %s stale keys and resolving %s unknown keys"
            , counts["created"], counts["deleted"], counts["removed"], counts["resolved"]
            )

        if any(usable) and usable[0].is_root_credentials():
            log.error("Can't programmatically rotate root credentials!")
            log.error("Try using IAM")
            self._changed = False
            return False

        if not any(usable):
            # All the keys will be removed when we save
            # Up to code after that to complain there is nothing to create new keys with
            for key in for_deletion:
                key.delete()
        else:
            if len(for_deletion) == 2 or (for_deletion and to_remain):
                # Make sure we have room for creating a new key
                deleting = for_deletion.pop(0)
                deleting.delete()
                usable = [pair for pair in usable if pair is not deleting]

            if for_rotation or not to_remain:
                if len(to_remain) > 1:
                    oldest = sorted(to_remain, key=lambda k: k.create_epoch)[0]
                    oldest.delete()
                    usable = [pair for pair in usable if pair is not oldest]

                iam_pair = usable[0].create_new()
                if half_life is None:
                    half_life = ask_user_for_half_life(access_key)
                iam_pair.set_half_life(half_life)
                self.add_key(iam_pair)
                usable.append(iam_pair)

            for key in for_deletion:
                # Delete the other keys marked for deletion
                key.delete()
                usable = [pair for pair in usable if pair is not key]

        usable_access_keys = [pair.aws_access_key_id for pair in usable]
        self.keys = [key for key in self.keys if key.iam_pair and key.iam_pair.aws_access_key_id in usable_access_keys]
        return True




from credo.asker import ask_user_for_secrets, ask_for_choice_or_new, ask_for_env, ask_user_for_saml, get_response, ask_for_choice
from credo.errors import CantEncrypt, CantSign, BadCredential, ProgrammerError, SamlNotAuthorized, CredoError
from credo.helper import print_list_of_tuples, make_export_commands, normalise_half_life
from credo.structure.credentials import SamlCredentials
from credo.amazon import IamPair, IamSaml
from credo.server import Server
from credo import structure

from textwrap import dedent
import pkg_resources
import platform
import requests
import logging
import base64
import pickle
import shutil
import json
import sys
import os

log = logging.getLogger("credo.actions")

def do_serve(credo, port=80, host="169.254.169.254", **kwargs):
    Server(host, port, credo).start()

def do_switch(credo, port=80, host="169.254.169.254", **kwargs):
    url = "http://{0}:{1}/latest/meta-data/switch/".format(host, port)
    chosen = credo._chosen = credo.make_chosen(rotate=False)
    if not isinstance(chosen, SamlCredentials):
        raise CredoError("Switch only supports idp roles")

    request = {"credentials": chosen}
    while True:
        response = requests.post(url, data=pickle.dumps(request), headers={"Content-Type": "application/octet-stream"})
        if response.status_code == 500:
            error = response.json()["error"]
            if error in ("NEED_AUTH", "BAD_PASSWORD"):
                if error == "BAD_PASSWORD":
                    log.error("Password was incorrect")
                password = get_response("Password for idp user {0}".format(chosen.keys.idp_username), password=True)
                request["basic_auth"] = base64.b64encode("{0}:{1}".format(chosen.keys.idp_username, password).encode('utf-8'))
            else:
                break
        else:
            break

    print("{0}: {1}".format(response.status_code, response.text))

def do_current(credo, **kwargs):
    """Print out what user is currently in our environment"""
    if "AWS_ACCESS_KEY_ID" not in os.environ or "AWS_SECRET_ACCESS_KEY" not in os.environ:
        print("There are currently no credentials in your environment!")
    else:
        iam_pair = IamPair.from_environment()
        print("Asking amazon for details")
        if not iam_pair.works:
            print("Your current credentials are not valid....")
        else:
            aliases = iam_pair.ask_amazon_for_account_aliases()
            if not aliases:
                aliases = ["<no_account_alias>"]
            print("You are currently \"{0}\" from \"{1}\" (account {2})".format(
                iam_pair.ask_amazon_for_username(), aliases[0], iam_pair.ask_amazon_for_account()
                ))

def do_unset(credo, **kwargs):
    """Just print out the exports and unsets necessary to unset credo exports"""
    exports = []
    for key, val in os.environ.items():
        if key.startswith("CREDO_UNSET"):
            name = key[12:]
            exports.append((name, val))
            exports.append((key, "CREDO_UNSET"))

    for command in make_export_commands(exports):
        print(command)

def do_exports(credo, chosen=None, repository=None, half_life=None, **kwargs):
    """Just print out the chosen creds"""
    half_life = normalise_half_life(half_life or getattr(credo, "half_life", None))

    if chosen is None:
        chosen = credo._chosen = credo.make_chosen(rotate=True, half_life=half_life)

    if repository is None:
        repository = credo.chosen.credential_path.repository

    shell_exports = {}
    for key, val in chosen.shell_exports():
        shell_exports[key] = val

    if not shell_exports:
        print("# {0} has no environment variables".format(chosen.path))

    unsetters = {}
    for key, val in shell_exports.items():
        name = "CREDO_UNSET_{0}".format(key)
        if key not in os.environ or os.environ[key] == val:
            unsetters[name] = "CREDO_UNSET"
        else:
            unsetters[name] = os.environ[key]

    print("## Find values to unset")
    do_unset(credo)

    print("\n## Values to be set now")
    for command in make_export_commands(sorted(shell_exports.items())):
        print(command)

    print("\n## And unsetters for when we change")
    for command in make_export_commands(sorted(unsetters.items()), no_transform=True):
        print(command)

    repository.synchronize()

def do_capture(credo, env=None, remove_env=None, all_accounts=False, all_users=False, find_user=False, **kwargs):
    """Capture environment variables"""
    part = credo.find_credential_path_part(all_accounts=all_accounts, all_users=all_users, find_user=find_user)
    if hasattr(part, "credential_path"):
        repository = part.credential_path.repository
    else:
        repository = part

    env, remove_env = ask_for_env(part, env, remove_env, ask_for_more=False)
    keys = part.get_env_file(credo.crypto)
    if keys.changed:
        repository.add_change("Capturing environment variables", [part.environment_location])

def do_env(credo, all_accounts=False, all_users=False, find_user=False, **kwargs):
    """display environment variables only"""
    part = credo.find_credential_path_part(all_accounts=all_accounts, all_users=all_users, find_user=find_user)
    if hasattr(part, "credential_path"):
        repository = part.credential_path.repository
    else:
        repository = part
    do_exports(credo, chosen=part, repository=repository)

def do_synchronize(credo, **kwargs):
    """Just synchronize some repo"""
    repo_name, location = credo.find_one_repository(want_new=False)
    structure.repository.synchronize(repo_name, location, credo.crypto)

def do_exec(credo, command, half_life=None, **kwargs):
    """Exec some command with aws credentials in the environment"""
    half_life = normalise_half_life(half_life or getattr(credo, "half_life", None))
    credo._chosen = credo.make_chosen(rotate=True, half_life=half_life)

    environment = dict(os.environ)
    for key, val in credo.chosen.shell_exports():
        if val == "CREDO_UNSET":
            if key in environment:
                del environment[key]
        else:
            environment[key] = val
    os.execvpe(command[0], command, environment)
    credo.chosen.credential_path.repository.synchronize()

def do_rotate(credo, force=False, half_life=None, **kwargs):
    """Rotate some keys"""
    log.info("Doing a rotation")
    half_life = normalise_half_life(half_life or getattr(credo, "half_life", None))
    credo.make_chosen(rotate=True, invalidate_creds=force, half_life=half_life).credential_path.repository.synchronize()

def do_remote(credo, remote=None, version_with=None, **kwargs):
    """Setup remotes for some repository"""
    repo_name, location = credo.find_one_repository()
    structure.repository.configure(repo_name, location, credo.crypto, new_remote=remote, version_with=version_with)

def do_import(credo, source=False, half_life=None, **kwargs):
    """Import some creds"""
    typ, info = ask_user_for_secrets(credo, source=source)
    if typ == "amazon":
        access_key, secret_key = info
        iam_pair = IamPair(access_key, secret_key, half_life=half_life)

        if not iam_pair.works:
            raise BadCredential("The credentials you just provided don't work....")
        half_life = normalise_half_life(half_life, access_key) or getattr(credo, "half_life", None)
        iam_pair.set_half_life(half_life)
        iam_pair._changed = False

    elif typ == "saml":
        access_key = None

        # Keep asking for username and password until they give one
        while True:
            idp_username = get_response("Idp username", default=os.environ.get("USER"))
            idp_password = get_response("Idp password", password=True)
            iam_pair = IamSaml(info, idp_username, idp_password, half_life=half_life)
            try:
                arns = iam_pair.arns
            except SamlNotAuthorized:
                continue

            saml_account = ask_for_choice("Which account do you want to use?", choices=sorted(arns, key=lambda a: a.role_arn))
            break
    else:
        raise ProgrammerError("Unknown credential type {0}".format(typ))

    structure, chains = credo.find_credentials(asker=ask_for_choice_or_new, want_new=True)
    creds = list(credo.credentials_from(structure, chains, typ=typ))[0]
    cred_path = creds.credential_path
    log.info("Making credentials for\trepo=%s\taccount=%s\tuser=%s", cred_path.repository.name, cred_path.account.name, cred_path.user.name)

    if typ != "saml":
        cred_path.repository.pub_key_syncer.sync()
        log.debug("Crypto has private keys %s", credo.crypto.private_key_fingerprints)
        log.debug("Crypto has public_keys %s", credo.crypto.public_key_fingerprints)

        if not credo.crypto.can_encrypt:
            raise CantEncrypt("No public keys to encrypt with", repo=cred_path.repository.name)
        if not credo.crypto.can_sign:
            log.error("Couldn't find any private keys matching your known public keys!")
            cred_path.repository.pub_key_syncer.sync(ask_anyway=True)
            if not credo.crypto.can_sign:
                raise CantSign("No private keys with matching public keys to sign with", repo=cred_path.repository.name)

    if typ == "amazon":
        account_id = cred_path.account.account_id(iam_pair=iam_pair)
        if iam_pair.ask_amazon_for_account() != account_id:
            raise BadCredential("The credentials you are importing are for a different account"
                , credentials_account_id=iam_pair.ask_amazon_for_account(), importing_into_account_name=cred_path.account.name, importing_into_account_id=account_id
                )

        username = cred_path.user.username(iam_pair=iam_pair)
        if iam_pair.ask_amazon_for_username() != username:
            raise BadCredential("The credentials you are importing are for a different user"
                , credentials_user=iam_pair.ask_amazon_for_username(), importing_into_user=username
                )

        creds.keys.add(iam_pair)
    else:
        contents = creds.contents
        contents.keys["provider"] = info
        contents.keys["role"] = saml_account.encrypted_values()
        contents.keys["idp_username"] = idp_username

    creds.save(half_life=half_life)
    cred_path.repository.synchronize()

def do_register_saml(credo, provider=None, **kwargs):
    """Register a saml provider"""
    if provider:
        credo.register_saml_provider(provider)
    else:
        ask_user_for_saml(credo)

def do_output_extension(credo, output, **kwargs):
    """Output the Chrome extension for the metadata magic server."""
    source = pkg_resources.resource_filename("credo", "ext")
    try:
        shutil.copytree(source, output)
    except OSError as error:
        raise CredoError("Failed to copy directory", source=source, output=output, error=error)

    print(dedent("""
        Congratulations, you know have the extension.
        - Please go into Chrome.
        - Go to chrome://extensions.
        - Enable developer mode.
        - Load unpacked extension.
            - Choose {0}
    """.format(output)))

def do_print_shell_function(credo, virtualenv=None, **kwargs):
    """Print the shell function to add to your environment."""

    if virtualenv is None:
        virtualenv = os.path.abspath(
                os.path.join( os.path.dirname(sys.argv[0]), "../")
            )

    mac_setup = """
        addr="169.254.169.254";
        loopback_interface="lo0";
        if ! ifconfig ${loopback_interface} | grep ${addr} > /dev/null; then;
            echo "creating $addr alias";
            sudo ifconfig lo0 alias $addr;
            plist=/Library/LaunchDaemons/delfick.credo.fake_metadata.plist;
            for action in unload load; do;
                sudo launchctl $action $plist;
            done;
        fi;
    """

    linux_setup = """
        addr="169.254.169.254";
        if ! ip route get $addr | grep "dev lo"; then
            echo "creating $addr alias";
            sudo ip addr add $addr dev lo
        fi;
    """

    if os.name == "posix":
        if platform.system() == "Darwin":
            setup = mac_setup
        else:
            setup = linux_setup
    else:
        setup = ""

    setup = "\n    ".join(line for line in setup.split('\n'))

    print(dedent("""
        Add the following to your shrc file (~/.bashrc, ~/.zshrc)

        ======================================================
        credo () {{
            {1}
            if {0}/bin/credo sourceable $@; then
                output=$({0}/bin/credo $@);
                if (($? == 0)); then
                    echo "$output" > /tmp/lolz;
                    source /tmp/lolz;
                else
                    echo "$output";
                fi;
            else
                {0}/bin/credo $@;
            fi
        }}
        switch () {{
            if [[ -z $1 ]]; then
                credo switch
            else
                credo --account $1 switch
            fi
        }}
        ======================================================
    """.format(virtualenv, setup)))

def do_create_launch_daemon(credo, virtualenv=None, **kwargs):
    """Write the LaunchConfig plist file."""
    output_file = "/Library/LaunchDaemons/delfick.credo.fake_metadata.plist"
    config_location = os.path.expanduser("~/.credo/config.json")

    if virtualenv == None:
        virtualenv = os.path.abspath(
                os.path.join( os.path.dirname(sys.argv[0]), "../")
            )

    if os.path.exists(output_file):
        print("File already exists! ({0})".format(output_file))
        return

    with open(output_file, "w") as fle:
        fle.write(dedent("""
            <?xml version="1.0" encoding="UTF-8"?>
            <!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
            <plist version="1.0">
            <dict>
                    <key>Label</key>
                    <string>delfick.credo.fake_metadata</string>
                    <key>ProgramArguments</key>
                    <array>
                            <string>{0}/bin/credo</string>
                            <string>--config</string>
                            <string>{1}</string>
                            <string>serve</string>
                    </array>
                    <key>RunAtLoad</key>
                    <true/>
                    <key>UserName</key>
                    <string>root</string>
                    <key>EnvironmentVariables</key>
                    <dict>
                        <key>HOME</key>
                        <string>{2}</string>
                    </dict>
                    <key>StandardOutPath</key>
                    <string>/var/log/credo/out.log</string>
                    <key>StandardErrorPath</key>
                    <string>/var/log/credo/err.log</string>
            </dict>
            </plist>
        """.format(virtualenv, config_location, os.path.expanduser("~"))))

    print(dedent("""
        LaunchDaemon has been written to {0}.

        To load the daemon use the following command:

           $ launchctl load {0}

        If there are every any problems, try:

           $ launchctl unload {0}
           $ launchctl load {0}

    """.format(output_file)))

def do_showavailable(credo, force_show_all=False, collapse_if_one=True, **kwargs):
    """Show all what available repos, accounts and users we have"""
    structure, chains = credo.find_credentials(no_mask=force_show_all)

    fltrs = ()
    if not force_show_all:
        fltrs = [(key, val) for key, val in [("repo", credo.repo), ("account", credo.account), ("user", credo.user)] if val]
    print_list_of_tuples(fltrs, "Using the filters")

    headings = ["Repositories", "Accounts", "Users"]

    def chains_to_dict(structure, chains):
        """Turn the list of chains into a dictionary"""
        dct = {}
        for chain in chains:
            rest, last = chain[1:-1], chain[-1]
            d = dct
            for part in rest:
                if part not in d:
                    d[part] = {}
                d = d[part]
            d[last] = lambda: list(credo.credentials_from(structure, [chain], complain_if_missing=True))[0]
        return dct

    def get_displayable(root, headings, indent="", underline_chain=None, sofar=None):
        """
        Return a structure for printing out our available credentials
        return as (heading, children)

        Where children is
            {child: (heading, grandchildren), child2: (heading, grandchildren)}
        """
        if sofar is None:
            sofar = []

        if underline_chain is None:
            underline_chain = ["="]

        def get_indented(s, prefix=""):
            """Print the string with leading indentation"""
            return "{0}{1}{2}".format(indent, prefix, s)

        def get_underlined(s, underline):
            """Get the indented str with an indented underline"""
            if underline:
                return "{0}\n{1}".format(get_indented(s), get_indented(underline * len(s)))
            else:
                return "{0}:".format(get_indented(s, ">> "))

        if not headings:
            return root
        else:
            heading = headings.pop(0)

            heading_underline = None
            if underline_chain:
                heading_underline = underline_chain.pop(0)

            children = {}
            for key, val in root.items():
                indented_key = get_indented(key)
                children[indented_key] = get_displayable(val, list(headings), indent + "    ", list(underline_chain), list(sofar) + [indented_key])

            if not children:
                return None
            return get_underlined(heading, heading_underline), children

    def display_creds(cred, indent=""):
        """Display info about the creds"""
        if callable(cred):
            cred = cred()
        as_string = cred.as_string()
        for line in as_string.split('\n'):
            print("{0}{1}".format(indent * 3, line))

    def display_result(result):
        """Display the result from get_displayable"""
        heading, children = result
        print("")
        print(heading)
        for child, values in children.items():
            print("")
            print(child)
            if isinstance(values, list) or isinstance(values, tuple) or isinstance(values, dict):
                display_result(values)
            elif values:
                display_creds(values, "    ")

    # Complain if no found chains
    if not chains:
        print("Didn't find any credential files")
        return

    # Special case if we only found one
    if collapse_if_one and len(chains) is 1:
        creds = list(credo.credentials_from(structure, chains, complain_if_missing=True))[0]
        cred_path = creds.credential_path
        fltr = [("repo", cred_path.repository.name), ("account", cred_path.account.name), ("user", cred_path.user.name)]
        print_list_of_tuples(fltr, "Only found one set of credentials")
        display_creds(creds)
        return

    # Or just do them all if found more than one
    root = chains_to_dict(structure, chains)
    result = get_displayable(root, headings)
    display_result(result)




from credo.errors import UserQuit, GitError, ProgrammerError
from credo.asker import ask_for_choice
from credo.versioning.base import Base

from contextlib import contextmanager
import tempfile
import logging
import shutil
import os

try:
    # Optional dependency is optional
    import pygit2
except ImportError:
    pygit2 = None

log = logging.getLogger("credo.versioning.git")

class GitDriver(Base):
    """Knows how to git"""
    def __init__(self, location):
        self.location = location

    @property
    def repo(self):
        """Get a repository object"""
        if not hasattr(self, "_repo"):
            self._repo = pygit2.Repository(self.location)
        return self._repo

    def synchronize(self, override=False):
        """Pull in changes from our remote"""
        if not self.remote:
            return

        self.resolve_dirty_repo()
        if self.repo.status():
            raise ProgrammerError("Somehow there are still changes in the repo....")

        origin = self.origin
        res = origin.fetch()
        if any(res.values()):
            log.info("Pulled in changes from %s", origin.url)
            oid = self.repo.lookup_reference("refs/remotes/origin/master").target

            log.info("Resetting %s to origin/master (%s)", self.location, oid)
            self.repo.reset(oid, pygit2.GIT_RESET_HARD)
        else:
            try:
                origin.push("refs/heads/master")
            except ValueError as error:
                log.error("Failed to push to remote repository\tremote=%s\terror=%s", origin.url, error)

    def resolve_dirty_repo(self):
        """Wait for the user to resolve any changes already in the repo"""
        while True:
            if not self.repo.status():
                break

            quit_choice = "Quit"
            fixed_choice = "I fixed it, please continue"
            choice = ask_for_choice("Seems there is already changes\tlocation={0}".format(self.location), choices=[quit_choice, fixed_choice])
            if choice == quit_choice:
                raise UserQuit()
            elif self.repo.status():
                log.error("Ummm, there's still changes already in the repository...\tlocation=%s", self.location)

    def determine_remote(self):
        """Get us back the url of the origin remote"""
        if self.origin:
            return self.origin.url

    @property
    def origin(self):
        for remote in self.repo.remotes:
            if remote.name == "origin":
                return remote

    def set_origin(self, new_remote):
        """Set our origin remote to be the new remote"""
        for remote in self.repo.remotes:
            if remote.name == "origin":
                remote.url = new_remote
                return

        self.repo.create_remote("origin", new_remote)

    def get_git_user(self):
        """Get us a user for our git"""
        name = self.repo.config.get_multivar("user.name")
        if name:
            name = name[0]
        else:
            name = "Credo"

        email = self.repo.config.get_multivar("user.email")
        if email:
            email = email[0]
        else:
            email = "credo@nowhereinparticular.com"

        return pygit2.Signature(name, email)

    def current_commit(self):
        """Return the current commit"""
        head = self.repo.lookup_reference("HEAD")
        try:
            return [head.resolve().target]
        except KeyError:
            return []

    def add_change(self, message, changed_files):
        """Add a change to the repo"""
        if not changed_files:
            log.warning("Told to add a change, but no files were specified\tmessage=%s", message)
        elif changed_files is True:
            self.repo.index.add_all([])
        else:
            for filename in changed_files:
                self.repo.index.add(filename)

        self.repo.index.write()
        if any(status in (pygit2.GIT_STATUS_INDEX_DELETED, pygit2.GIT_STATUS_INDEX_MODIFIED, pygit2.GIT_STATUS_INDEX_NEW) for status in self.repo.status().values()):
            oid = self.repo.index.write_tree()
            author = committer = self.get_git_user()
            self.repo.create_commit('HEAD', author, committer, message, oid, self.current_commit())

    def is_versioned(self):
        """Yes we are versioned"""
        return True

    def deleteme(self):
        """Delete the versioning!"""
        quit_choice = "Quit"
        confirm_choice = "Yes I do want to delete!"

        git_folder = os.path.join(self.location, ".git")
        choice = ask_for_choice("Are you sure you want to remove {0}?".format(git_folder), [quit_choice, confirm_choice])
        if choice == quit_choice:
            raise UserQuit()
        else:
            shutil.rmtree(git_folder)

    def initialise(self, new_remote=None):
        """Setup the .git folder, with optional remote"""
        log.info("Setting up a .git folder")
        git_folder = os.path.join(self.location, ".git")
        if os.path.exists(git_folder):
            raise GitError("Trying to initialise git, but .git folder already exists", location=self.location)

        pygit2.init_repository(self.location)
        self.add_change("Initial commit", changed_files=True)
        if new_remote:
            self.change_remote(new_remote)

    def change_remote(self, new_remote):
        """Setup our new remote!"""
        if new_remote is None and self.remote:
            log.info("Removing current remote (%s)", self.remote)
        elif self.remote:
            if self.remote != new_remote:
                log.info("Changing remote to %s from %s", new_remote, self.remote)
            else:
                log.info("Remote for this repository is already %s", new_remote)
        elif new_remote:
            log.info("Adding remote\tlocation=%s\tnew origin=%s", self.location, new_remote)

        if new_remote is None:
            if self.remote:
                self.repo.remote_remote("origin")
        elif new_remote != self.remote:
            with self.temp_clone(new_remote):
                self.set_origin(new_remote)
                self.remote = new_remote

    @contextmanager
    def temp_clone(self, url):
        """Clone a url into a temporary place, yield that place, then delete it"""
        tmpdir = None
        try:
            tmpdir = tempfile.mkdtemp()
            repo = os.path.join(tmpdir, "repo")
            try:
                pygit2.clone_repository(url, repo)
            except pygit2.GitError as error:
                raise GitError("Couldn't clone the new remote", url=url, repo=self.location, error_type=error.__class__.__name__, error=error)
            yield repo
        finally:
            if os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)

